﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.IteratedSymbolicExpressionConstruction {
  [StorableClass]
  [Item("BreathFirstSymbolicExpressionConstructionPolicy", "Deterministic (almost) breath-first search")]
  // simulates breath first search by following branches in the search tree that have been visited least often
  public class BreathFirstSymbolicExpressionConstructionPolicy : SymbolicExpressionConstructionPolicyBase {

    public BreathFirstSymbolicExpressionConstructionPolicy()
      : base() {
    }

    private readonly Dictionary<string, int> tries = new Dictionary<string, int>();

    public override void Initialize(SymbolicExpressionTreeProblem problem, IRandom random) {
      base.Initialize(problem, random);
      tries.Clear();
    }

    protected override int Select(IReadOnlyList<object> followStates, IRandom random) {
      // we want to find the action sequence that has been tried least frequently so far

      Contract.Assert(followStates.Any());
      var states = followStates.Cast<string>().ToArray();

      // find action with minimum tries so far 
      var bestIdx = 0;
      var bestS = states.First();
      int i = 0;
      int minTries = int.MaxValue;
      foreach (string s in states) {

        int sTries;
        tries.TryGetValue(s, out sTries);
        if (sTries < minTries) {
          bestS = s;
          minTries = sTries;
          bestIdx = i;
        }
        i++;
      }

      // increase tries of selected action
      if (!tries.ContainsKey(bestS)) {
        tries.Add(bestS, 1);
      } else {
        tries[bestS]++;
      }
      return bestIdx;

    }

    public sealed override void Update(IEnumerable<object> stateSequence, double quality) {
      // ignore
    }

    protected override object CreateState(ISymbolicExpressionTreeNode root, List<ISymbol> actionSequence, ISymbolicExpressionTreeNode parent, int childIdx) {
      // TODO more compact representation than a dictionary of strings (many duplicate prefixes)
      return string.Join(" ", actionSequence); // return the sequence of actions as string (=path in the search tree)
    }

    #region IItem
    protected BreathFirstSymbolicExpressionConstructionPolicy(BreathFirstSymbolicExpressionConstructionPolicy original, Cloner cloner)
      : base(original, cloner) {
    }

    [StorableConstructor]
    protected BreathFirstSymbolicExpressionConstructionPolicy(bool deserializing) : base(deserializing) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BreathFirstSymbolicExpressionConstructionPolicy(this, cloner);
    }

    #endregion
  }
}
